module Exercise08_SMT(
  SmtVar,
  IAExpr(..),
  Formula(Gt, BoolVar, Le, Equiv), true, false, disj, conj, atomic,
  SmtType(..),
  SmtEncoder, 
  runSmtEncoder,
  getNewSmtVariable,
  assertFormula) where

{- Task 2 (4 points)

Do not change the module declaration above.

Below you find some code that describes the layout of
an SMT encoder. This encoder has three features:

- it can store formulas that should be asserted
- it can generate fresh variables x1, x2, x3, ...
- it can be run, where eventually everything is exported into
  the common SMT-Lib 2 format
  (https://smt-lib.org)

The template already contains the definition of formulas and of the
rendering to SMT-Lib 2.

Define a suitable new type SmtEncoder and implement the missing 
methods. 

Important: whenever you generate a fresh variable, then internally 
  a SmtVarDecl has to be added.

Restriction: the usage of RWS is forbidden here, you should use 
  one or many more basic monads or monad transformers.

Hint: you might need further or other imports, depending on the monads
  you would like to use. 
-}

import Control.Monad.Reader

-- setting up the formula and the Smt datatypes

newtype SmtVar = SmtVar Int deriving Eq

instance Show SmtVar where
  show (SmtVar x) = "x" ++ show x

data IAExpr = IAVar SmtVar | IAConst Int 
  deriving (Eq)
  
instance Show IAExpr where
  show (IAVar x) = show x
  show (IAConst x) 
    | x < 0 = "(- " ++ show (- x) ++ ")"
    | otherwise = show x

data Formula = 
    Gt IAExpr IAExpr 
  | Le IAExpr IAExpr 
  | BoolVar SmtVar
  | Conj [Formula]
  | Disj [Formula]
  | Equiv Formula Formula
  deriving (Eq)
  
instance Show Formula where
  show (Gt a b) = "(> " ++ show a ++ " " ++ show b ++ ")"
  show (Le a b) = "(<= " ++ show a ++ " " ++ show b ++ ")"
  show (BoolVar b) = show b
  show (Equiv f g) = "(= " ++ show f ++ " " ++ show g ++ ")"
  show (Conj []) = "true"
  show (Conj fs) = "(and" ++ foldr (\ f s -> " " ++ show f ++ s) ")" fs
  show (Disj []) = "false"
  show (Disj fs) = "(or" ++ foldr (\ f s -> " " ++ show f ++ s) ")" fs

atomic :: Formula -> Bool
atomic (Disj []) = True
atomic (Conj []) = True
atomic (BoolVar _) = True
atomic (Gt _ _) = True
atomic (Le _ _) = True
atomic _ = False

-- smart constructors for the formula type
false = Disj []
true = Conj []

disj phis = let phis' = filter (/= false) phis
  in if true `elem` phis' then true else
    case phis' of 
      [phi] -> phi
      _ -> Disj phis'

conj phis = let phis' = filter (/= true) phis
  in if false `elem` phis' then false else
    case phis' of 
      [phi] -> phi
      _ -> Conj phis'

-- towards statements of the SMT-Lib format

data SmtType = SmtBool | SmtInt

instance Show SmtType where
  show SmtBool = "Bool"
  show SmtInt = "Int"

data SmtStmt = 
   SmtVarDecl SmtType SmtVar
 | SmtAssert Formula

instance Show SmtStmt where
  show (SmtVarDecl ty x) = "(declare-fun " ++ show x ++ " () " ++ show ty ++ ")"
  show (SmtAssert phi) = "(assert " ++ show phi ++ ")"

showSmt2 :: [SmtStmt] -> String
showSmt2 stmts = unlines $ header ++ map show stmts ++ footer 
  where header = ["(set-logic QF_LIA)"]
        footer = ["(check-sat)"]


--- The encoder (you have to change the code below + the imports)


-- One can hide a monad transformer stack in a newtype, using GeneralizedNewtypeDeriving
-- The example type "ReaderT String IO a" needs to be changed to something suitable!
newtype SmtEncoder a = SmtEncoder (ReaderT String IO a)
  deriving (Functor, Applicative, Monad, 
     MonadReader String, MonadIO)

-- Implement these three functions
assertFormula :: Formula -> SmtEncoder ()
assertFormula = undefined

getNewSmtVariable :: SmtType -> SmtEncoder SmtVar
getNewSmtVariable = undefined

runSmtEncoder :: SmtEncoder a -> String
runSmtEncoder = undefined "use" showSmt2 "and various runMonad(T)-functions"

{- 
after completion, 
running smtEncoderTest should result in the output that is similar to 
the following one, where you might have used a different numbering scheme, etc.

(set-logic QF_LIA)
(declare-fun x1 () Bool)
(declare-fun x2 () Bool)
(declare-fun x3 () Int)
(assert (or x1 (> x3 5)))
(assert (= x2 (or x1 (> x3 5))))
(check-sat)

-}

smtEncoderTest :: IO ()
smtEncoderTest = do
  let output = runSmtEncoder (do
       b <- getNewSmtVariable SmtBool
       c <- getNewSmtVariable SmtBool
       x <- getNewSmtVariable SmtInt
       let f1 = Gt (IAVar x) (IAConst 5)
       let f2 = disj [BoolVar b, f1]
       assertFormula f2
       assertFormula (Equiv (BoolVar c) f2))
  putStrLn output
  